#!/bin/bash
#
# Product setup script - Loki Entertainment Software
#
# Modifications by Borland/Inprise Corp.:
# 02/24/2000: added functions CheckKernel, CheckLibc, CheckX and enabling code in MAIN
#
# Version constants to check against
KERNEL_MAJOR=2
KERNEL_MINOR=2
KERNEL_BUILD=0
X_MAJOR=11
X_MINOR=0
LIBC_MAJOR=6
GLIBC_MAJOR=2
GLIBC_MINOR=1
GLIBC_BUILD=2
QT_MAJOR=2
JPEG_MAJOR=6
JPEG_MINOR=2
JPEG_BUILD=0
NCURSES_MAJOR=4

# BEGIN STRING TABLE
# These constants should be translated for international use
#
CHECK_DEP="Checking dependencies"
KERNEL_VER="Kernel version"
FAILED="FAILED"
OK="OK"
LIBC_VER="Glibc version"
X_VER="X11 Server"
QT_VER="Qt version"
LIBJPEG_VER="Libjpeg version"
CANNOT_FIND="Setup cannot locate"
X_SERVER="X Server"
X_FAILURE="Setup could not detect a suitable X11 server."
CONTINUE="Do you want to install Kylix (default Y) ? "
EXIST_INSTALL="Setup has detected an existing installation at this location:"
NO_RELOCATE1="If you wish to reinstall, you will not be able to choose a"
NO_RELOCATE2="different installation location."
LIBNCURSES="Libncurses version"
LOADER_ERR1="Setup has determined that your system needs an updated version of glibc"
LOADER_ERR2="(C runtime library). The version of glibc that is required varies by"
LOADER_ERR3="distribution. Please read the document \"PREINSTALL\" on this disc."
FATAL_ERR1="Your system does not meet the minimum system requirements."
FATAL_ERR2="Setup cannot continue."
KYLIX_RUNNING="Kylix is currently running. Please shut down Kylix and run the installer again."
VERS_ERR1="Setup has detected another edition of Kylix is already installed on this system."
VERS_ERR2="You must remove the other edition before installing this one."
# END STRING TABLE


function CheckKernel {

    major=`uname -r | awk -F. '{printf $1}'`
    minor=`uname -r | awk -F. '{printf $2}'`
    kernelOK=1
    if test $major -gt $KERNEL_MAJOR; then
	kernelOK=0
    elif test $major -eq $KERNEL_MAJOR; then
	if test $minor -ge $KERNEL_MINOR; then
	    kernelOK=0
	fi
    fi
    kernel_version="$major.$minor"
    return $kernelOK;
}

function CheckLibc {
    major=0
    minor=0
    build=0
    
    # copy the chklibc file to someplace where we know it can be executed
    chklibc="$HOME/.chklibc$$"
    if [ -e setup.data/chklibc ]; then
	cp "setup.data/chklibc" "$chklibc"
	chmod 700 "$chklibc"
	libcver=`$chklibc`
    else
	libc_version="cannot load chklibc"
	return 1
    fi

    if [ -e $chklibc ]; then
	rm -f $chklibc
    fi

    major=`echo $libcver |awk -F. '{printf $1}'`
    minor=`echo $libcver |awk -F. '{printf $2}'`
    build=`echo $libcver |awk -F. '{printf $3}'`
    libc_version="$major.$minor.$build"
#    if [ "$libc" = "glibc-2.1" -o "$libc" = "glibc-2.0" ]; then
    if [ $major -ge $GLIBC_MAJOR ]; then
#	libcfile=`ls /lib/libc-2.* 2> /dev/null | tail -1`
#	if [ -n "$libcfile" ]; then
#	    minor=`echo $libcfile |awk -F. '{printf $2}'`
#	    build=`echo $libcfile |awk -F. '{printf $3}'`
	    if [ $minor -gt $GLIBC_MINOR ]; then
		return 0;
	    elif [ $minor -eq $GLIBC_MINOR ]; then
		if [ $build -ge $GLIBC_BUILD ]; then
		    return 0;
		else
		    return 1;
		fi
	    else
		return 1;
	    fi
#	else 
#	    libc_version="could not determine"
#	    return 1;
#	fi
    else
	libc_version="less than 2.0"
	return 1;
    fi
}

function CheckLoader {
    if [ -n "`expr "$args" : '\(.*-g\)'`" ]; then
	ldresult=0
	return 0;
    else
	ldresult=1
	# copy the borpretest file to someplace where we know it can be executed
	chkldpath="$HOME/.borchkld$$"
	if [ -d borpretest ]; then
	    mkdir "$chkldpath"
	    cp borpretest/* "$chkldpath" &>/dev/null
	    chmod 700 "$chkldpath/borpretest"
	    if [ -n "$LD_LIBRARY_PATH" ]; then
		oldldpath=`echo $LD_LIBRARY_PATH`
		export LD_LIBRARY_PATH="$chkldpath":$LD_LIBRARY_PATH
		`$chkldpath/borpretest -q`
		ldresult=$?
		export LD_LIBRARY_PATH=$oldldpath
	    else
		export LD_LIBRARY_PATH="$chkldpath"
		`$chkldpath/borpretest -q`
		ldresult=$?
		unset LD_LIBRARY_PATH
	    fi

	    if [ -d "$chkldpath" ]; then
		rm -rf "$chkldpath"
	    fi
	
	    loader_result="$ldresult"
	    return $ldresult
	else
	    loader_result="Test program not found"
	    return 1
	fi
    fi
}

function CheckX {
    xOK=1
    major=0
    minor=0
    build=0
    # Since X may not be in their path, we have to find it
    #findx=`whereis xdpyinfo`
    findx=`whereis X | awk '{printf $2}'`
    if [ -n "$findx" ]; then
	xOK=0
	x_server="$findx"
    else
	x_server="could not find"
    fi
    return $xOK;
}

function CheckQt {
    if test $QTDIR=""; then
	return 1
    else
	return 0
    fi
}

function CheckJpeg {
    jpegOK=1
    major=0
    minor=0
    build=0
    if [ -e /usr/lib/libjpeg.so.62.0.0 ]; then
        jpegOK=0    
	jpeg_version="6.2.0"
    elif [ -L /usr/lib/libjpeg.so.6 ]; then
	jpeg=`ls /usr/lib/libjpeg.so.6.*`
	major=`echo $jpeg |awk -F. '{print $3}'`
	minor=`echo $jpeg |awk -F. '{print $4}'`
	build=`echo $jpeg |awk -F. '{print $5}'`
	if [ $major -gt $JPEG_MAJOR ]; then
	    jpegOK=0
	elif [ $major -eq $JPEG_MAJOR -a $minor -gt $JPEG_MINOR ]; then
	    jpegOK=0
	elif [ $major -eq $JPEG_MAJOR -a $minor -eq $JPEG_MINOR -a $build -ge $JPEG_BUILD ]; then
	    jpegOK=0
	fi
	jpeg_version="$major.$minor.$build"
    fi
    
    return $jpegOK
}

function CheckCurses {
    
    result=1

    # first copy chkcurses to someplace that is executable.
    chkcurses="$HOME/.chkcurses$$"
    if [ -e setup.data/chkcurses ]; then
	cp "setup.data/chkcurses" "$chkcurses"
	chmod 700 "$chkcurses"
	ncurses=`$chkcurses`
    else
	found_ncurses="cannot load chkcurses"
	return 1
    fi

    if [ $? -eq 0 ]; then
	found_ncurses="found via RPM query"
	result=0
    elif [ $? -eq 1 ]; then
	if [ -e /usr/lib/libncurses.so.4 ]; then
	    found_ncurses="/usr/lib/libncurses.so.4"
	    result=0
	elif [ -e /usr/local/lib/libncurses.so.4 ]; then
	    found_ncurses="/usr/local/lib/libncurses.so.4"
	    result=0
	elif [ -e /lib/libncurses.so.4 ]; then
	    found_ncurses="/lib/libncurses.so.4"
	    result=0
	fi
    fi

    if [ -e "$chkcurses" ]; then
	rm -f "$chkcurses"
    fi
    return $result
}

# Go to the proper setup directory (if not already there)
cd `dirname $0`

# Return the appropriate architecture string
function DetectARCH {
	status=1
	case `uname -m` in
		i?86)  echo "x86"
			status=0;;
		*)     echo "`uname -m`"
			status=0;;
	esac
	return $status
}

# Return the appropriate version string
function DetectLIBC {
      status=1
      if [ -f `echo /lib/libc.so.6* | tail -1` ]; then
	      if fgrep GLIBC_2.1 /lib/libc.so.6* 2>&1 >/dev/null; then
	              echo "glibc-2.1"
	              status=0
	      else    
	              echo "glibc-2.0"
	              status=0
	      fi        
      elif [ -f /lib/libc.so.5 ]; then
	      echo "libc5"
	      status=0
      else
	      echo "unknown"
      fi
      return $status
}

# Detect the Linux environment
arch=`DetectARCH`
libc=`DetectLIBC`

# Find the installation program
function try_run
{
    setup=$1
    shift
    fatal=$1
    if [ "$1" != "" ]; then
        shift
    fi

    # First find the binary we want to run
    failed=0
    setup_bin="setup.data/bin/$arch/$libc/$setup"
    if [ ! -f "$setup_bin" ]; then
        setup_bin="setup.data/bin/$arch/$setup"
        if [ ! -f "$setup_bin" ]; then
            failed=1
        fi
    fi
    if [ $failed -eq 1 ]; then
        if [ "$fatal" != "" ]; then
            cat <<__EOF__
This installation does not support $libc on $arch

Please contact Technical Support.
__EOF__
            exit 1
        fi
        return $failed
    fi

    # Try to run the binary
    # The executable is here but we can't execute it from CD
    setup="$HOME/.setup$$"
    cp "$setup_bin" "$setup"
    chmod 700 "$setup"
    if [ "$fatal" != "" ]; then
	if [ -n "$pathargs" ]; then
	    cat <<EOF

$EXIST_INSTALL

$pathargs

$NO_RELOCATE1
$NO_RELOCATE2

EOF
	    echo -n "$CONTINUE "
	    read response
	    case $response in
		n|N|no|No|NO )
		    fatal_error=1
		    ;;
		* )
		    "$setup" $installargs -i$pathargs $*
		    ;;
	    esac
	else
	    "$setup" $installargs $*
	fi
        failed=$?
    else
	if [ -n "$pathargs" ]; then
	    cat <<EOF

$EXIST_INSTALL

$pathargs

$NO_RELOCATE1
$NO_RELOCATE2

EOF
	    echo -n "$CONTINUE "
	    read response
	    case $response in
		n|N|no|No|NO )
		    fatal_error=1
		    ;;
		* )
		    "$setup" $installargs -i$pathargs $* 2>/dev/null
		    ;;
	    esac
	else
	    "$setup" $installargs $* 2>/dev/null
	fi
        failed=$?
    fi
    rm -f "$setup"
    return $failed
}

function CheckSystem {

    fatal_error=0
    error=""

#    if [ -n "`ps -ax |grep -i kylix`" ]; then
#	echo "$KYLIX_RUNNING"
#	echo "$FATAL_ERR2"
#	return 1
#    fi

    if ! CheckKernel; then
	fatal_error=1
	echo "$KERNEL_VER >= $KERNEL_MAJOR.$KERNEL_MINOR.$KERNEL_BUILD....$FAILED"
    else
	echo "$KERNEL_VER >= $KERNEL_MAJOR.$KERNEL_MINOR.$KERNEL_BUILD....$OK"
    fi

    if ! CheckLoader; then
	fatal_error=1
	echo "$LIBC_VER....$FAILED"
	cat <<EOF

$LOADER_ERR1
$LOADER_ERR2
$LOADER_ERR3

EOF
    elif ! CheckLibc; then
	fatal_error=1
	echo "$LIBC_VER >= $GLIBC_MAJOR.$GLIBC_MINOR.$GLIBC_BUILD....$FAILED"
    else
	echo "$LIBC_VER >= $GLIBC_MAJOR.$GLIBC_MINOR.$GLIBC_BUILD....$OK"
    fi

    if ! CheckX; then
	echo "$X_VER....$FAILED"
        cat <<EOF

$X_FAILURE

EOF
	echo -n "$CONTINUE "
	read response
	case $response in
        n|N|no|No|NO )
	  fatal_error=1
	  ;;
	* )
	  ;;
        esac
    else
	echo "$X_VER....$OK"
    fi

#    if ! CheckQt; then
#	fatal_error=0
#	echo "$QT_VER >= $QT_MAJOR....$FAILED"
#    else
#	echo "$QTVER >= $QT_MAJOR....$OK"
#    fi

    if ! CheckJpeg; then
	fatal_error=1
	echo "$LIBJPEG_VER >= $JPEG_MAJOR.$JPEG_MINOR.$JPEG_BUILD....$FAILED"
    else
	echo "$LIBJPEG_VER >= $JPEG_MAJOR.$JPEG_MINOR.$JPEG_BUILD....$OK"
    fi

#    if ! CheckCurses; then
#	fatal_error=1
#	echo "$LIBNCURSES = $NCURSES_MAJOR....$FAILED"
    #    else
#	echo "$LIBNCURSES = $NCURSES_MAJOR....$OK"
#    fi

    if test $fatal_error -eq 1; then
	cat <<EOF

$FATAL_ERR1
$FATAL_ERR2

EOF
	return 1
    fi
    echo 
    echo
    return 0
}

function CheckPreviousInstall {
    uid=`id -u`
    if [ "$uid" = "0" ]; then
	sh setup.data/chkprevvers.sh
	if [ $? -eq 0 ]; then
	    chkprev="$HOME/.chkprev$$"
	    cp "setup.data/borcheckprev" "$chkprev"
	    chmod 700 "$chkprev"
	    pathargs=`$chkprev 2>/dev/null`
	    if [ -e "$chkprev" ]; then
		rm -f "$chkprev"
	    fi
	else
	    cat <<EOF

$VERS_ERR1
$VERS_ERR2

$FATAL_ERR2
EOF
	    exit 1
	fi
    fi
    return 0
}

function CheckForCorel {
    uid=`id -u`
    if [ "$uid" = "0" ]; then
	dist=`grep -i "corel" /etc/issue`
	if [ -n "$dist" ]; then
	    found_corel="Corel Linux: $dist"
	    installargs="-m $installargs"
	else
	    found_corel="Not Corel Linux"
	fi
    else
	found_corel="non-root install...check not performed"
    fi
}

function WriteLog {
    findx=`whereis xdpyinfo | awk '{printf $2}'`
    if [ -n "$findx" ]; then
    cat <<EOF > ~/borland_kylix_install_log
Borland Kylix install log -- `date`
OS: Linux
kernel version: $kernel_version
Corel Linux check: $found_corel
loader test result code: $loader_result
libc version: $libc_version
libjpeg version: $jpeg_version
xdpyinfo:
`$findx 2>&1`
EOF
    else
	cat <<EOF > ~/borland_kylix_install_log
Borland Kylix install log -- `date`
OS: Linux
kernel version: $kernel_version
Corel Linux check: $found_corel
libc version: $libc_version
libjpeg version: $jpeg_version
#found libncurses: $found_ncurses
xdpyinfo: could not locate xdpyinfo
EOF

    fi

}

fatal_error=0

# Initial display
echo
echo
echo
echo "                                   BORLAND KYLIX"
echo 
echo "$CHECK_DEP..."

kernel_version=""
libc_version=""
x_server=""
jpeg_version=""
found_corel=""
found_ncurses=""
loader_result=""
args=$*

if ! CheckSystem; then
    exit 1
fi

pathargs=""
if [ -n "`expr "$args" : '\(.*-m\)'`" ]; then
    installargs="-m"
else
    installargs="" 
fi
CheckForCorel
WriteLog
if CheckPreviousInstall; then
    # Try to run the setup program
    status=0
    rm -f "$setup"
    if ! try_run setup.gtk; then
	if [ $fatal_error -eq 0 ]; then
	    if ! try_run setup -fatal; then
		echo "The setup program seems to have failed on $arch/$libc"
		echo
		echo "Please contact Technical Support."
		status=1
	    fi
	fi
    fi
fi
exit $status



























