{ *************************************************************************** }
{                                                                             }
{ Kylix and Delphi Cross-Platform Visual Component Library                    }
{                                                                             }
{ Copyright (c) 2000, 2001 Borland Software Corporation                       }
{                                                                             }
{ This file may be distributed and/or modified under the terms of the GNU     }
{ General Public License (GPL) version 2 as published by the Free Software    }
{ Software Foundation and appearing at http://www.borland.com/kylix/gpl.html. }
{                                                                             }
{ Licensees holding a valid Borland No-Nonsense License for this Software may }
{ use this file in accordance with such license, which appears in the file    }
{ license.txt that came with this Software.                                   }
{                                                                             }
{ *************************************************************************** }


unit DirSel;

{R-,T-,H+,X+}

interface

uses
  SysUtils, Types, Classes, QGraphics, QControls, QForms, QDialogs,
  QStdCtrls, QComCtrls, QExtCtrls, QImgList;

type
  TDirSelDlg = class(TForm)
    Panel1: TPanel;
    Button1: TButton;
    Button2: TButton;
    TV: TTreeView;
    ImageList1: TImageList;
    Panel2: TPanel;
    Panel3: TPanel;
    Label1: TLabel;
    Panel4: TPanel;
    procedure TVItemDoubleClick(Sender: TObject; Node: TTreeNode);
    procedure TVExpanded(Sender: TObject; Node: TTreeNode);
    procedure FormCreate(Sender: TObject);
    procedure TVCustomDrawItem(Sender: TCustomViewControl;
      Item: TCustomViewItem; Canvas: TCanvas; Rect: TRect;
      State: TCustomDrawState; Stage: TCustomDrawStage;
      var DefaultDraw: Boolean);
  private
    FRootDir: string;
    FDir: string;
    procedure AddDirectories(Node: TTreeNode; Dir: string);
    function GetAbsolutePath(Node: TTreeNode): string;
    procedure SetDir(const Value: string);
    procedure SetRootDir(const Value: string);
  public
    function SelectedDir: string;
    property Directory: string read FDir write SetDir;
    property RootDirectory: string read FRootDir write SetRootDir;
  end;

implementation

{$R *.xfm}

{$IFDEF LINUX}
uses 
  Libc;

var
  FCurrentDir: string = '';

function SelectProc(const p1: PDirEnt): Integer; cdecl;
var
  StatBuf: TStatBuf;
  S: string;
begin
  Result := 0;
  if StrLen(p1^.d_name) = 0 then Exit;
  S := FCurrentDir + PathDelim + p1^.d_name;
  if stat(PChar(S), StatBuf) = 0 then
    Result := (StatBuf.st_mode and S_IFDIR) * Cardinal((Ord(p1^.d_name[0]) - Ord('.')));
end;

function CompareProc(const p1, p2: Pointer): Integer; cdecl;
begin
  Result := AnsiCompareStr(PDirEnt(p1^)^.d_name, PDirEnt(p2^)^.d_name);
end;

function HasSubDirs(const Dir: string): Boolean;
var
  NumEntries: Integer;
  NameList,
  Next: PPDirEnt;
  I: Integer;
begin
  Result := False;
  if Dir <> '' then
  begin
    FCurrentDir := Dir;
    NumEntries := scandir(PChar(Dir), NameList, @SelectProc, nil);
    Result := NumEntries > 0;
    Next := NameList;
    for I := 0 to NumEntries - 1 do
    begin
      Libc.free(Next^);
      Inc(Next);
    end;
    if NumEntries > -1 then Libc.Free(NameList);
  end;
end;
{$ENDIF}

procedure TDirSelDlg.AddDirectories(Node: TTreeNode; Dir: string);
{$IFDEF LINUX}
var
  DirInfo: PDirEnt;
  NumEntries: Integer;
  NameList,
  Next: PPDirEnt;
  wet: wordexp_t;
  PP: PPChar;
  I: Integer;
  NewNode: TTreeNode;
begin
  if Dir <> '' then
  begin
    if wordexp(PChar(Dir), wet, 0) = 0 then
    try
      PP := wet.we_wordv;
      Dir := '';
      for I := 0 to wet.we_wordc - 1 do
      begin
        Dir := Dir + PP^;
        Inc(PP);
        if I < wet.we_wordc - 1 then
          Dir := Dir + ' ';
      end;
    finally
      wordfree(wet);
    end;

    if (Dir <> PathDelim) and (Dir[Length(Dir)] = PathDelim) then
      System.Delete(Dir, Length(Dir), 1);
    FCurrentDir := Dir;
    NumEntries := scandir(PChar(Dir), NameList, @SelectProc,
      @CompareProc);
    Next := NameList;
    for I := 0 to NumEntries - 1 do
    begin
      DirInfo := Next^;
      NewNode := TV.Items.AddChild(Node, DirInfo^.d_name);
      NewNode.ImageIndex := 0;
      NewNode.HasChildren := HasSubDirs(Dir + PathDelim + NewNode.Text);
      Libc.free(DirInfo);
      Inc(Next);
    end;
    if NumEntries > -1 then Libc.free(NameList);
    if Node.Level = 0 then Node.Text := Dir;
  end;
end;
{$ENDIF}
{$IFDEF MSWINDOWS}
begin
end;
{$ENDIF}


procedure TDirSelDlg.SetRootDir(const Value: string);
{$IFDEF LINUX}
var
  RootNode: TTreeNode;
begin
  TV.Items.Clear;
  FRootDir := Value;
  if FRootDir = '' then FRootDir := PathDelim;
  if (FRootDir <> PathDelim) and (FRootDir[Length(FRootDir)] = PathDelim) then
    System.Delete(FRootDir, Length(FRootDir), 1);
  RootNode := TV.Items.Add(nil, FRootDir);
  RootNode.ImageIndex := 0;
  AddDirectories(RootNode, FRootDir);
  TV.Selected := RootNode;
end;
{$ENDIF}
{$IFDEF MSWINDOWS}
begin
end;
{$ENDIF}

function TDirSelDlg.GetAbsolutePath(Node: TTreeNode): string;
{$IFDEF LINUX}
begin
  while Node <> nil do
  begin
    Result := Node.Text + PathDelim + Result;
    Node := Node.Parent;
  end;
end;
{$ENDIF}
{$IFDEF MSWINDOWS}
begin
end;
{$ENDIF}

procedure TDirSelDlg.TVItemDoubleClick(Sender: TObject; Node: TTreeNode);
{$IFDEF LINUX}
begin
  if Node.Count = 0 then
    AddDirectories(Node, GetAbsolutePath(Node));
end;
{$ENDIF}
{$IFDEF MSWINDOWS}
begin
end;
{$ENDIF}


procedure TDirSelDlg.TVExpanded(Sender: TObject; Node: TTreeNode);
{$IFDEF LINUX}
begin
  TVItemDoubleClick(Sender, Node);
end;
{$ENDIF}
{$IFDEF MSWINDOWS}
begin
end;
{$ENDIF}

procedure TDirSelDlg.SetDir(const Value: string);
{$IFDEF LINUX}
var
  StartDir: string;
  Node: TTreeNode;
  I, P: Integer;
  SubDir: PChar;
begin
  FDir := Value;
  StartDir := Value;
  if TV.Items.Count = 0 then Exit;
  P := Pos(FRootDir, StartDir);
  if P = 1 then
    Delete(StartDir, P, Length(FRootDir));
  for I := 1 to Length(StartDir) do
    if (StartDir[I] = PathDelim) then StartDir[I] := #0;
  SubDir := PChar(StartDir);
  if SubDir[0] = #0 then
    SubDir := @SubDir[1];
  Node := TV.Items.GetFirstNode;
  while SubDir[0] <> #0 do
  begin
    Node := Node.GetFirstChild;
    while (Node <> nil) and (AnsiCompareText(Node.Text, SubDir) <> 0) do
      Node := Node.GetNextSibling;
    if Node = nil then Break
    else begin
      Node.Expand(False);
      Node.Selected := True;
    end;
    SubDir := SubDir + StrLen(SubDir) + 1;
  end;
  TV.Selected.MakeVisible(False);
end;
{$ENDIF}
{$IFDEF MSWINDOWS}
begin
end;
{$ENDIF}

procedure TDirSelDlg.FormCreate(Sender: TObject);
{$IFDEF LINUX}
begin
  TV.HandleNeeded;
end;
{$ENDIF}
{$IFDEF MSWINDOWS}
begin
end;
{$ENDIF}

procedure TDirSelDlg.TVCustomDrawItem(Sender: TCustomViewControl;
  Item: TCustomViewItem; Canvas: TCanvas; Rect: TRect;
  State: TCustomDrawState; Stage: TCustomDrawStage;
  var DefaultDraw: Boolean);
{$IFDEF LINUX}
var
  Node: TTreeNode;
begin
  DefaultDraw := True;
  Node := TTreeNode(Item);
  if Node.Selected then Node.ImageIndex := 1 else Node.ImageIndex := 0;
end;
{$ENDIF}
{$IFDEF MSWINDOWS}
begin
end;
{$ENDIF}

function TDirSelDlg.SelectedDir: string;
{$IFDEF LINUX}
begin
  Result := '';
  if TV.Selected <> nil then
  begin
    Result := GetAbsolutePath(TV.Selected);
    Delete(Result, 1, 1);
  end;
end;
{$ENDIF}
{$IFDEF MSWINDOWS}
begin
end;
{$ENDIF}

end.
