/* ************************************************************************* *
 *                                                                           *
 * delphiuw.c --- Delphi Unwinder                                            *
 *                                                                           *
 * Copyright (c) 1994,2001 Borland Software Corporation                      *
 *                                                                           *
 * This file may be distributed under the terms of the GNU General Public    *
 * License (GPL) version 2 as published by the Free Software Foundation and  *
 * appearing at http://www.borland.com/kylix/gpl.html.                       *
 *                                                                           *
 * Licensees holding a valid Borland No-Nonsense License for this Software   *
 * may use this file in accordance with such license, which appears in the   *
 * file license.txt that came with this software.                            *
 *                                                                           *
 * ************************************************************************ */


#include	<stdio.h>
#include	<string.h>
#include	<stdlib.h>
#include	<assert.h>

#include	"diag.h"
#include	"arith.h"
#include	"unwind.h"
#include	"unwindfi.h"

// Turn this off for shipping.  See objout.c in dcc32 for more info.
//#define	GEN_EXC_NAMES

#ifdef	DO_STATISTICS
static	int				RunningCodeSizeTotal;
static	int				RunningExcSizeTotal;
static	int				RunningUnwindSizeTotal;
static	int				RunningRangeSizeTotal;
static  int				RunningProcCountTotal;
static	int				RunningExcProcTotal;
static	int				RunningFramedProcTotal;
static	int				RunningPotentialSavings;
static	int				RunningNonFramedProcUSize;
static	int				RunningSmallProcCount;
static	int				RunningSmallProcRanges;
#endif

#ifdef	__linux__
typedef unsigned long DWORD;
#endif

static void *DecodeFrameInfo(unsigned char *EndOfProcedure, ExceptionFrameInfo *FrameInfo)
{
	unsigned char	Buffer[ARITH_MAX_COMPRESSED_UNSIGNED_SIZE * 6 + 1];
	unsigned char *	p;
	int				i;
	unsigned char *	Start;

	/*
		The frame info is compressed at the end of the unwind info and
		code ranges.  It is reversed, too, since we don't know the size
		of the info.  So we have to copy the bytes, reverse them and
		decompress them.  We copy enough bytes to make sure that we got
		all the compressed data.  This is safe, since there is always
		enough data on the other side of the frame info to make up for
		overruns.
	*/
	p = EndOfProcedure - sizeof(Buffer);
	i = sizeof(Buffer);
	while	(i)
		Buffer[--i] = *p++;
	Start = p = Buffer;

	memset(FrameInfo, 0, sizeof(*FrameInfo));

	/*
		The flags byte is guaranteed.  Everything else is conditional.
	*/
	FrameInfo->Flags = *p++;

	if	(FrameInfo->Flags & UFI_F_CALLEEPOP)
		FrameInfo->CalleePopAmount = Arith_RdUnsigned(&p);

	if	(FrameInfo->Flags & UFI_F_SROFFS)
	{
		if	(FrameInfo->Flags & UFI_F_HASFRAME)
			FrameInfo->SavedRegsEBPOffs = Arith_RdUnsigned(&p);
		else
			FrameInfo->OffsetToPush = Arith_RdUnsigned(&p);
	}

	if	(FrameInfo->Flags & UFI_F_HASRANGES)
		FrameInfo->RangeCount = Arith_RdUnsigned(&p);

	if	(!(FrameInfo->Flags & UFI_F_HASFRAME) || (FrameInfo->Flags & UFI_F_NESTEDHANDLER))
	{
		FrameInfo->UnwindInfoSize = Arith_RdUnsigned(&p);
		FrameInfo->ProcLen = Arith_RdUnsigned(&p);
	}
	else if	(FrameInfo->Flags & UFI_F_HASRANGES)
	{
		FrameInfo->ProcLen = Arith_RdUnsigned(&p);
	}
	return (void *)(EndOfProcedure - (p - Start));
}

#ifdef	DO_STATISTICS
static	void SplitName(const char *Name, char *BaseName, char *MemberName)
{
	char *	Member;
	Member = strchr(Name, '.');
	if	(Member)
	{
		strncpy(BaseName, Name, Member - Name);
		BaseName[Member - Name] = 0;
		strcpy(MemberName, Member + 1);
	}
	else
	{
		BaseName[0] = 0;
		strcpy(MemberName, Name);
	}
}

static void DumpStatistics(void *Context)
{
	int				UnitCount;
	PCMapUnit *		Units;
	int				i;
	char *			TextStart;
	PCMapHeader *	Header;
	int				CodeSizeTotal;
	int				ExcSizeTotal;
	int				UnwindSizeTotal;
	int				RangeSizeTotal;
	int				ProcCountTotal;
	char			LastBaseName[4096];
	char			BaseName[4096];
	char			MemberName[4096];

	LastBaseName[0] = 0;
	BaseName[0] = 0;

	Header = (PCMapHeader *)Context;

	TextStart = (char *)Context - Header->CodeSegmentSize;
	printf("Image Range: %08x, %08x\n", TextStart, Context);
	CodeSizeTotal = ExcSizeTotal = UnwindSizeTotal = RangeSizeTotal = 0;
	ProcCountTotal = 0;
	
	UnitCount = Header->UnitCount;
	Units = (PCMapUnit *)(Header + 1);
	printf("%-25s %-4s %-4s/%-7s (%-4s/%-7s, %-4s/%-7s)\n",
			"Procedure", "Code ", "Exc ", "%", "Unwd", "%", "Rngs", "%");
	for	(i = 0; i < UnitCount; i++)
	{
		PCMapProc *	Procs;
		int			ProcCount;
		int			j;
		Procs = (PCMapProc *)(((char *)Context) + Units[i].ProcInfoOffs);
		ProcCount = Units[i].ProcCount;
		ProcCountTotal += ProcCount;
		RunningProcCountTotal += ProcCount;
		for	(j = 0; j < ProcCount; j++)
		{
			ExceptionFrameInfo 		FrameInfo;
			int *					pNameLen;
			char *					Name;
			int						ExcRangeSize;
			int						ExcInfoSize;
			unsigned char *			EndOfUnwindInfo;
			unsigned char *			End;
			int						ProcLen;
			
			End = (unsigned char *)(TextStart + Procs[j].End);
			EndOfUnwindInfo = DecodeFrameInfo(End, &FrameInfo);

			pNameLen = ((int *)EndOfUnwindInfo) - 1;
			Name = ((char *)pNameLen) - *pNameLen;
			SplitName(Name, BaseName, MemberName);
			if	(strcmp(BaseName, LastBaseName))
			{
				strcpy(LastBaseName, BaseName);
				if	(*BaseName)
					printf("%s\n", BaseName);
			}
			if	 (*BaseName)
				strcpy(BaseName, "  ");
			//      name procsize exc size/% (unwind size/%, range size/%)
			ExcRangeSize = FrameInfo.RangeCount * sizeof(ExceptionRangeOffsets);
			RunningExcProcTotal += FrameInfo.RangeCount;

			if	(!FrameInfo.RangeCount && (FrameInfo.Flags & UFI_F_HASFRAME))
				RunningFramedProcTotal++;
			if	(!FrameInfo.RangeCount && !(FrameInfo.Flags & UFI_F_HASFRAME))
				RunningNonFramedProcUSize += FrameInfo.UnwindInfoSize;
			if	(FrameInfo.RangeCount && (FrameInfo.Flags & UFI_F_HASFRAME))
				RunningPotentialSavings += FrameInfo.UnwindInfoSize;

			ExcInfoSize = (End - EndOfUnwindInfo) + FrameInfo.UnwindInfoSize + ExcRangeSize;

			if	(FrameInfo.ProcLen)
				ProcLen = FrameInfo.ProcLen;
			else
				ProcLen = Procs[j].End - Procs[j].Start - ExcInfoSize;
			if	(ProcLen <= 256 && FrameInfo.RangeCount)
			{
				RunningSmallProcCount++;
				RunningSmallProcRanges += FrameInfo.RangeCount;
			}

			CodeSizeTotal += ProcLen;
			ExcSizeTotal += ExcInfoSize;
			UnwindSizeTotal += FrameInfo.UnwindInfoSize;
			RangeSizeTotal += ExcRangeSize;
			
			RunningCodeSizeTotal += ProcLen;
			RunningExcSizeTotal += ExcInfoSize;
			RunningUnwindSizeTotal += FrameInfo.UnwindInfoSize;
			RunningRangeSizeTotal += ExcRangeSize;
			
			printf("%s%-*s %4d %4d/%5.2f (%4d/%5.2f, %4d/%5.2f)\n",
				   BaseName,
				   *BaseName ? 23 : 25,
				   MemberName,
				   ProcLen,
				   ExcInfoSize,
				   (float)ExcInfoSize / (float)ProcLen,
				   FrameInfo.UnwindInfoSize,
				   (float)FrameInfo.UnwindInfoSize / (float)ExcInfoSize * 100.0f,
				   ExcRangeSize,
				   (float)ExcRangeSize / (float)ExcInfoSize * 100.0f);
		}
	}
	printf("Totals: C(%d) E(%d) (%4.2f), U(%d) R(%d) (%4.2f, %4.2f), PC(%d)\n",
	       CodeSizeTotal,
	       ExcSizeTotal,
	       (float)ExcSizeTotal / (float)CodeSizeTotal,
	       UnwindSizeTotal,
	       RangeSizeTotal,
	       (float)UnwindSizeTotal / (float)ExcSizeTotal,
	       (float)RangeSizeTotal / (float)ExcSizeTotal,
	       ProcCountTotal);
	printf("RTotals: C(%d) E(%d) (%4.2f), U(%d) R(%d) (%4.2f, %4.2f), PC(%d)\n",
	       RunningCodeSizeTotal,
	       RunningExcSizeTotal,
	       (float)RunningExcSizeTotal / (float)RunningCodeSizeTotal,
	       RunningUnwindSizeTotal,
	       RunningRangeSizeTotal,
	       (float)RunningUnwindSizeTotal / (float)RunningExcSizeTotal,
	       (float)RunningRangeSizeTotal / (float)RunningExcSizeTotal,
	       RunningProcCountTotal);
	printf("Procs with exception handling: %d (%4.2f%% of total)\n",
		   RunningExcProcTotal, (float)RunningExcProcTotal / (float)RunningProcCountTotal * 100.0f);
	printf("Procs without exception handling, but with frames: %d (%4.2f%% of total)\n",
		   RunningFramedProcTotal, (float)RunningFramedProcTotal / (float)RunningProcCountTotal * 100.0f);
	printf("Procs without exception handling, no frames, unwind size: %d (%4.2f%% of total)\n",
		   RunningNonFramedProcUSize, (float)RunningNonFramedProcUSize / (float)RunningCodeSizeTotal * 100.0f);
	printf("Potential unwind space savings: %d (%4.2f%% of total => new percentage of %4.2f%%)\n",
			RunningPotentialSavings,
			(float)RunningPotentialSavings / (float)RunningExcSizeTotal * 100.0f,
			(float)(RunningExcSizeTotal - RunningPotentialSavings) / (float)RunningCodeSizeTotal * 100.0f);
	printf("Small procs with exception handling: %d (%d total ranges)\n",
		   RunningSmallProcCount, RunningSmallProcRanges);
}
#endif

static	ExceptionRangeOffsets *	FindRange(
	unsigned long 			ProcRelOffs,
	ExceptionRangeOffsets *	Ranges,
	int						RangeCount)
{
	int	i;

	/*
		Note the shift of the inclusion fence here:  The end is included,
		and the start is excluded.  This may cause a problem for OS exception
		dispatching if the faulting instruction is a one byte instruction at
		the end of the code range.
	*/
	for	(i = 0; i < RangeCount; i++)
		if	(ProcRelOffs > Ranges[i].Start && ProcRelOffs <= Ranges[i].End)
			return &Ranges[i];

	return NULL;
}

typedef	struct	DelphiDecoder
{
	unsigned char *	UnwindInfo;
	unsigned char *	EndUnwindInfo;
	int				DeltaESP;
	int				DeltaOffs;
	unsigned long	CurOffs;
	unsigned long	ProcStart;		//debugging
}	DelphiDecoder;

static	void	InitDelphiDecoder(
	DelphiDecoder *	Decoder,
	unsigned char *	Info,
	int				InfoSize,
	unsigned long ProcStart)
{
	Decoder->UnwindInfo = Info;
	Decoder->EndUnwindInfo = Info + InfoSize;
	Decoder->DeltaESP = 0;
	Decoder->DeltaOffs = 0;
	Decoder->CurOffs = 0;
	Decoder->ProcStart = ProcStart;
}

static	int	DecodeTo(
	DelphiDecoder *	Decoder,
	unsigned long TargetOffs)
{
	int	Res;
	int	DeltaESP;
	int	DeltaOffs;
	unsigned long	NewOffs;

	Res = 0;
	DeltaESP = Decoder->DeltaESP;
	DeltaOffs = Decoder->DeltaOffs;
	NewOffs = Decoder->CurOffs;
	DEBUG_CONTINUE("unwind.frame.decode", ("DecodeTo:\n"));
	while	(NewOffs + DeltaOffs <= TargetOffs &&
			 Decoder->UnwindInfo < Decoder->EndUnwindInfo)
	{
		NewOffs += DeltaOffs;
		Res += DeltaESP;
		DeltaESP = Arith_RdSigned(&Decoder->UnwindInfo);
		DeltaOffs = Arith_RdUnsigned(&Decoder->UnwindInfo);
		DEBUG_CONTINUE("unwind.frame.decode", ("%08x (%4d): DeltaEsp = %d\n",
											Decoder->ProcStart + NewOffs,
											NewOffs,
											Res));
	}
	Decoder->DeltaESP = DeltaESP;
	Decoder->DeltaOffs = DeltaOffs;
	Decoder->CurOffs = NewOffs;
	return Res;
}

static	int	BitCount(unsigned int Bits)
{
	int	Count = 0;

	while	(Bits)
	{
		Count += Bits & 1;
		Bits = Bits >> 1;
	}
	return Count;
}

/*
	The two structures below are borroed from unwind.c.  This is improper, but this file
	was split from unwind.c late in the game.  The proper thing to do is to expand the SetGR/GetGR
	interface in unwind.c's implementation to accept the needed registers, and then we go
	through the APIs in this file.
*/
#ifdef	__IA32__
#define MAX_EXTENSION     512
typedef unsigned long	ULONG;
typedef unsigned char	UBYTE;
#define SIZE_FP_REGISTERS      80
typedef struct FP_SAVE
{
    ULONG   Control;
    ULONG   Status;
    ULONG   Tag;
    ULONG   ErrorOff;
    ULONG   ErrorSel;
    ULONG   DataOff;
    ULONG   DataSel;
    UBYTE   Registers[SIZE_FP_REGISTERS];
    ULONG   Cr0NpxState;
}	FP_SAVE;

/*
	Don't rearrange _Unwind_Context arbitrarily!  There's ASM code here which
	manipulates this directly.
*/
struct	_Unwind_Context
{
    ULONG   Eip;
    ULONG   Ebp;
    ULONG   Esp;
    ULONG   Ebx;
    ULONG   Dr0;
    ULONG   Dr1;
    ULONG   Dr2;
    ULONG   Dr3;
    ULONG   Dr6;
    ULONG   Dr7;
    FP_SAVE FloatSave;
    ULONG   SegGs;
    ULONG   SegFs;
    ULONG   SegEs;
    ULONG   SegDs;
    ULONG   Edi;
    ULONG   Esi;
    ULONG   Edx;
    ULONG   Ecx;
    ULONG   Eax;
    ULONG   SegCs;
    ULONG   EFlags;
    ULONG   SegSs;
    UBYTE   ExtendedRegisters[MAX_EXTENSION];
};
#endif

static	void	UnwindDelphiFrame(
	_Unwind_Context *	Context,
	unsigned char *		Info,
	ExceptionFrameInfo *FrameInfo,
	unsigned long 		ProcRelOffs,
	unsigned long		ProcStartAddr,
	unsigned long		StopAddr)
{
	unsigned long	PushRegsOffs;
	unsigned char	FrameFlags;
	int				ESPToPush;
	int				ESPToStopAddr;
	int				ESPToEIP;
	int				SavedRegs;
	DelphiDecoder	Decoder;
	int				CalleePopAmount;

	FrameFlags = FrameInfo->Flags;


#pragma message ("Use context APIs instead of direct manipulation")
	DEBUG_PRINT("unwind.frame.info", ("UnwindDelphiFrame (to %p)\n", StopAddr));
	DEBUG_CONTINUE("unwind.frame.info", ("    Unwind Info Size: %d bytes.", FrameInfo->UnwindInfoSize));
	DEBUG_CODE("unwind.frame.info", {
		DEBUG_CLASS("unwind.frame.info");
		if	(FrameFlags & UFI_F_HASFRAME)
		{
			DEBUG_CONTINUE_(("  EBP frame present\n"));
		}
		else
		{
			DEBUG_CONTINUE_(("  No frame present\n"));	
		}
				});
	DEBUG_CONTINUE("unwind.frame.info", ("    Pushed regs:"));
	SavedRegs = FrameInfo->Flags & (UFI_F_REG_EBX |
									UFI_F_REG_ESI |
									UFI_F_REG_EDI |
									UFI_F_REG_EBP);

	DEBUG_CODE("unwind.frame.info", {
		DEBUG_CLASS("unwind.frame.info");
		if	(SavedRegs)
		{
			if	(FrameFlags & UFI_F_REG_EBX) DEBUG_CONTINUE_((" EBX"));
			if	(FrameFlags & UFI_F_REG_ESI) DEBUG_CONTINUE_((" ESI"));
			if	(FrameFlags & UFI_F_REG_EDI) DEBUG_CONTINUE_((" EDI"));
			if	(FrameFlags & UFI_F_REG_EBP) DEBUG_CONTINUE_((" EBP"));
			if	(FrameInfo->Flags & UFI_F_HASFRAME)
			{
				DEBUG_CONTINUE_(("\n    EBP rel offs to saved regs: %08x\n", FrameInfo->SavedRegsEBPOffs));
			}
			else
			{
				DEBUG_CONTINUE_(("\n    Offset to push %08x\n", FrameInfo->OffsetToPush));
			}
		}
		else
		{
			DEBUG_CONTINUE_((" None\n"));
		}
		});

	CalleePopAmount = FrameInfo->CalleePopAmount;
	DEBUG_CONTINUE("unwind.frame.info", ("    Callee pops %d bytes.\n", CalleePopAmount));

	/*
		Under some circumstances, there is no unwind information whatsoever
		present.  This happens if there is a BP frame.  Also, if we have a frame,
		and we want to unwind this frame completely, we don't have to bother
		decoding the unwind info, we can just ditch the stack frame
		immediately.  If we have UFI_F_NESTEDHANDLER, and we're doing a partial
		unwind, we have to use the unwind info, because otherwise we might toss
		a crucial return address for a finally block if we just unwind the stack
		back to the saved regs push point.
	*/

#pragma message ("Optimize this code next time round")
/*
	We're not properly testing here to see if we really need the unwind info.
	We can still optimize the case where there is unwind info, but we are completely
	unwinding the frame, and there is an EBP frame.  Be careful, though - SavedRegs
	will have UFI_F_REG_EBP set in this case, which isn't really right, because
	it's overloaded to UFI_F_NESTEDHANDLER.
*/
	if	((FrameInfo->Flags & (UFI_F_HASFRAME | UFI_F_NESTEDHANDLER)) == UFI_F_HASFRAME)
	{
		/*
			We need to retrieve EBX, ESI and EDI, if necessary.  If we're
			unwinding the function completely, we toss the whole frame,
			otherwise we just reset ESP to the end of the prologue sequence.
		*/
		if	(!StopAddr)
		{
			if	(SavedRegs)
			{
				DWORD *	SavedRegsPtr;
	
				SavedRegsPtr = (DWORD *)(Context->Ebp + FrameInfo->SavedRegsEBPOffs);
				
				if	(SavedRegs & UFI_F_REG_EDI)
					Context->Edi = *SavedRegsPtr++;
				if	(SavedRegs & UFI_F_REG_ESI)
					Context->Esi = *SavedRegsPtr++;
				if	(SavedRegs & UFI_F_REG_EBX)
					Context->Ebx = *SavedRegsPtr++;
			}
			Context->Eip = *(DWORD *)(Context->Ebp + 4);
			Context->Esp = Context->Ebp + 4 + 4 + FrameInfo->CalleePopAmount;
			Context->Ebp = *((DWORD *)Context->Ebp);
		}
		else
		{
			DEBUG_CONTINUE("unwind.frame.info", ("    Resetting ESP to prologue (%p - %x - %x).\n", Context->Ebp, -FrameInfo->SavedRegsEBPOffs, BitCount(SavedRegs) * 4));
			Context->Esp = Context->Ebp + FrameInfo->SavedRegsEBPOffs - BitCount(SavedRegs) * 4;
		}
		return;
	}

	PushRegsOffs = FrameInfo->OffsetToPush;

	assert(!SavedRegs || (ProcRelOffs > PushRegsOffs));

	InitDelphiDecoder(&Decoder, Info, FrameInfo->UnwindInfoSize, ProcStartAddr);

	if	(SavedRegs)
	{
		DWORD *	SavedRegsPtr;

		/*
			Start munching the offset info until we pass the pushed regs.
		*/
		ESPToPush = DecodeTo(&Decoder, PushRegsOffs);

		/*
			If we're only unwinding partially, then we take the
			further step of finding the change in the stack across the
			entire range up to the stopping point provided us.
		*/
		if	(PushRegsOffs != StopAddr)
			ESPToStopAddr = DecodeTo(&Decoder, StopAddr) + ESPToPush;
		else
			ESPToStopAddr = ESPToPush;

		/*
			Now we read until we get to the current execution pointer,
			and we'll have all the stack mods we need.
		*/
		ESPToEIP = DecodeTo(&Decoder, ProcRelOffs);

		/*
			Blow away the stack to the point where entered the protected range,
			then restore the saved regs, and do the rest of the cleanup.
		*/
		if	(!StopAddr)
		{
			SavedRegsPtr = (DWORD *)(Context->Esp - (ESPToEIP + (ESPToStopAddr - ESPToPush)));
			if	((SavedRegs & UFI_F_REG_EBP) && !(FrameInfo->Flags & UFI_F_HASFRAME))
				Context->Ebp = *SavedRegsPtr++;
			if	(SavedRegs & UFI_F_REG_EDI)
				Context->Edi = *SavedRegsPtr++;
			if	(SavedRegs & UFI_F_REG_ESI)
				Context->Esi = *SavedRegsPtr++;
			if	(SavedRegs & UFI_F_REG_EBX)
				Context->Ebx = *SavedRegsPtr++;
			Context->Esp -= ESPToStopAddr + ESPToEIP;
			Context->Eip = *((DWORD *)Context->Esp);
			Context->Esp += 4 + CalleePopAmount; // ditch call and any stack parms
			if	(FrameFlags & UFI_F_HASFRAME)
				Context->Ebp = *((DWORD *)Context->Ebp);
		}
		else
		{
			Context->Esp -= ESPToEIP;
		}
	}
	else
	{
		if	(StopAddr)
		{
			ESPToStopAddr = DecodeTo(&Decoder, StopAddr);
			ESPToEIP = DecodeTo(&Decoder, ProcRelOffs);
			Context->Esp -= ESPToEIP;
		}
		else
		{
			/*
				No saved regs.  We just gather ESP mods.
			*/
			ESPToEIP = DecodeTo(&Decoder, ProcRelOffs);
	
			/*
				Got the ESP mods.  Now we undo them in the context, and
				update the EIP.
			*/
			Context->Esp -= ESPToEIP;
			Context->Eip = *((DWORD *)Context->Esp);
			Context->Esp += 4 + CalleePopAmount; // ditch call and any stack parms
			if	(FrameFlags & UFI_F_HASFRAME)
				Context->Ebp = *((DWORD *)Context->Ebp);
		}
	}
}

static	_Unwind_Reason_Code	UNWIND_CC DelphiPersonalityFn(
	int version,
	_Unwind_Action actions,
	_Unwind_Exception *exceptionObject,
	_Unwind_Context *context,
	void *ImplementationData)
{
	unsigned long *			BPChain;
	int						ExcInfoSize;
	ExceptionFrameInfo 		FrameInfo;
	unsigned char *			End;
	unsigned char *			EndOfUnwindInfo;
	unsigned long			ProcRelOffs;
	unsigned long			ProcStartAddr;
	int						i;
	ExceptionRangeOffsets *	Ranges;
	ExceptionRangeOffsets *	Range;
	unsigned char *			UnwindInfo;
	_Unwind_Context 		NewContext;
#ifdef	GEN_EXC_NAMES
	int *					pNameLen;
	char *					Name;
#endif

	NewContext = *context;
	End = (unsigned char *)ImplementationData;
	EndOfUnwindInfo = DecodeFrameInfo(End, &FrameInfo);
#ifdef	GEN_EXC_NAMES
	pNameLen = ((int *)EndOfUnwindInfo) - 1;
	Name = ((char *)pNameLen) - *pNameLen;

	DEBUG_PRINT("unwind.control.phases", ("  Found %s", Name));
#else
	DEBUG_PRINT("unwind.control.phases", ("  Found %p", End - FrameInfo.ProcLen));
#endif

#ifdef	GEN_EXC_NAMES
	ProcStartAddr = ((unsigned long)EndOfUnwindInfo) - FrameInfo.ProcLen - 4 - *pNameLen -
									sizeof(ExceptionRangeOffsets) * FrameInfo.RangeCount -
									FrameInfo.UnwindInfoSize;
#else
	ProcStartAddr = ((unsigned long)EndOfUnwindInfo) - FrameInfo.ProcLen -
									sizeof(ExceptionRangeOffsets) * FrameInfo.RangeCount -
									FrameInfo.UnwindInfoSize;
#endif
	ProcRelOffs = _BorUnwind_GetIP(context) - ProcStartAddr;
#ifdef	GEN_EXC_NAMES
	Ranges = (ExceptionRangeOffsets *)((char *)EndOfUnwindInfo - *pNameLen - 4 - 
									sizeof(ExceptionRangeOffsets) * FrameInfo.RangeCount);
#else
	Ranges = (ExceptionRangeOffsets *)((char *)EndOfUnwindInfo -
									sizeof(ExceptionRangeOffsets) * FrameInfo.RangeCount);
#endif
	UnwindInfo = (unsigned char *)Ranges - FrameInfo.UnwindInfoSize;

	BPChain = (unsigned long *)_BorUnwind_GetGR(context, _UW_REG_EBP);
	if	(actions & _UA_SEARCH_PHASE)
	{
		/*
			Now we have to dig down into the local frame's ranges, and 
			see which part of the fn really wants this thing.
		*/
		Range = FindRange(ProcRelOffs, Ranges, FrameInfo.RangeCount);
		if	(Range)
		{
			/*
				Someone here wants the exception.  Fine.
			*/
			DEBUG_CONTINUE("unwind.control.phases",
						(" - I want the exception - %08x in [%08x, %08x)\n",
						_BorUnwind_GetIP(context),
						ProcStartAddr + Range->Start,
						ProcStartAddr + Range->End));
			return _URC_HANDLER_FOUND;
		}

		DEBUG_CONTINUE("unwind.control.phases", ("\n"));
	}
	if	(actions & _UA_CLEANUP_PHASE)
	{
		if	(actions & _UA_HANDLER_FRAME)
		{
			/*
				Unwind the current frame, back to the start of the current range
			*/
			Range = FindRange(ProcRelOffs, Ranges, FrameInfo.RangeCount);
			assert(Range);

			DEBUG_CONTINUE("unwind.control.phases", (" - transfer control now\n"));
			UnwindDelphiFrame(context,
							  UnwindInfo,
							  &FrameInfo,
							  ProcRelOffs,
							  ProcStartAddr,
							  Range->Start);
			/*
				Got the handler.  Install it.
			*/
			_BorUnwind_SetIP(context, ProcStartAddr + Range->Handler);
			_BorUnwind_SetGR(context, _UW_REG_EBX, _BorUnwind_FindGOT(_BorUnwind_GetIP(context)));
			DEBUG_PRINT("unwind.control.phases",
						("Transferring control with new EIP = %08x, ESP = %08x, EBP = %08x\n",
						context->Eip, context->Esp, context->Ebp));
			return _URC_INSTALL_CONTEXT;
		}
		DEBUG_CONTINUE("unwind.control.phases", ("\n"));
	}

	DEBUG_PRINT("unwind.control.phases",
				("Entered unwind with        EIP = %08x, ESP = %08x, EBP = %08x\n",
				context->Eip, context->Esp, context->Ebp));
	UnwindDelphiFrame(context,
					  UnwindInfo,
					  &FrameInfo,
					  ProcRelOffs,
					  ProcStartAddr,
					  0);
	DEBUG_PRINT("unwind.control.phases",
				("Continuing unwind with new EIP = %08x, ESP = %08x, EBP = %08x\n",
				context->Eip, context->Esp, context->Ebp));
	return _URC_CONTINUE_UNWIND;
}

UNWIND_PUBLIC __personality_routine	UNWIND_CC _BorUnwind_DelphiLookup(
	_Unwind_codePtr Addr,
	void *			Context,
	void **			pImplementationData)
{
	int				UnitCount;
	PCMapUnit *		Units;
	int				i;
	char *			TextStart;
	long			AddrOffs;
	PCMapHeader *	Header;

	Header = (PCMapHeader *)Context;

if	(Header->Signature != PCMAPHEADER_SIGNATURE)
{
	printf("FATAL: version mismatch in unwind.so.  Maybe you need to update the compiler/unwind.so?\n");
	assert(!"Have to fail here");
}

	TextStart = (char *)Context - Header->CodeSegmentSize;
	UnitCount = Header->UnitCount;
	Units = (PCMapUnit *)(Header + 1);
	AddrOffs = Addr - (unsigned long)TextStart;
	for	(i = 0; i < UnitCount; i++)
	{
		if	(AddrOffs >= Units[i].Start && AddrOffs < Units[i].End)
		{
			// Found it.  Now find the proc.
			PCMapProc *	Procs;
			int			ProcCount;
			int			j;
			Procs = (PCMapProc *)(((char *)Context) + Units[i].ProcInfoOffs);
			ProcCount = Units[i].ProcCount;
			for	(j = 0; j < ProcCount; j++)
			{
				if	(AddrOffs >= Procs[j].Start && AddrOffs < Procs[j].End)
				{
					// Found the proc.
					if	(pImplementationData)
						*pImplementationData = (void *)(TextStart + Procs[j].End);
					return DelphiPersonalityFn;
				}
			}
		}
	}
	return NULL;
}

#ifdef	INTERNAL_DEBUGGING
UNWIND_PUBLIC void	UNWIND_CC _BorUnwind_CrawlStack(void)
{
	__personality_routine	Personality;
	void *					PersonalityData;
	_Unwind_Context			Context;
	_Unwind_RegVal			CatchIP;

	GetContext(&Context);

	DEBUG_PRINT("unwind.control.crawl", ("CrawlStack\n"));

	for	(;;)
	{
		ExceptionFrameInfo 		FrameInfo;
		unsigned char *			End;
		unsigned char *			EndOfUnwindInfo;
		int *					pNameLen;
		char *					Name;

		CatchIP = _BorUnwind_GetIP(&Context);
		Personality = FindPersonality(&Context, CatchIP, &PersonalityData, 0);
		if	(!Personality)
		{
			DEBUG_PRINT("unwind.control.crawl", (" Unable to find personality\n"));
			return;
		}

		EndOfUnwindInfo = DecodeFrameInfo((unsigned char *)PersonalityData, &FrameInfo);
		pNameLen = ((int *)EndOfUnwindInfo) - 1;
		Name = ((char *)pNameLen) - *pNameLen;
		DEBUG_PRINT("unwind.control.crawl", ("%s\n", Name));

		if ((Personality)(_UNWIND_VERSION,
					  _UA_SEARCH_PHASE,
					  NULL,
					  &Context,
					  PersonalityData) == _URC_HANDLER_FOUND)
		{
			DEBUG_PRINT("unwind.control.crawl", ("--- Handler found; end of trace ---\n"));
			return;
		}
	}
}
#endif

