/* **************************************************************************
 *                                                                          *
 * unwind.c --- Exception stack unwind routines as specified by the C++ ABI *
 *   for IA-64:Exception Handling specification                             *
 *                                                                          *
 * Copyright (c) 2001 Borland Software Corporation.                         *
 *                                                                          *
 * This file may be distributed and/or modified under the terms of the GNU  *
 * Lesser General Public License (LGPL) version 2.1 as published by the     *
 * Free Software Foundation and appearing at                                *
 * http://www.borland.com/kylix/lgpl.html.                                  *
 *                                                                          *
 * Licensees holding a valid Borland No-Nonsense License for this Software  *
 * may  use this file in accordance with such license, which appears in the *
 * file license.txt that came with this Software.                           *
 *                                                                          *
 * *************************************************************************/

#include	<stdio.h>
#include	<string.h>
#include	<stdlib.h>
#include	<assert.h>
#ifdef	__linux__
#include	<signal.h>
#endif

#include	"diag.h"
#include	"arith.h"
#include	"unwind.h"

/*
	The following debug classes are used here:

"unwind.control.phases"
"unwind.search.image"
"unwind.search.verbose"
"unwind.frame.decode"
"unwind.frame.info"
*/

typedef	struct	LookupList
{
	_Unwind_IPLookupFn	Func;
	_Unwind_codePtr		Start;
	_Unwind_codePtr		End;
	_Unwind_codePtr		GOT;
	void *				Context;
	struct LookupList *	Next;
	struct LookupList *	Prev;
}	LookupList;

static	LookupList *	LookList;


UNWIND_PUBLIC int UNWIND_CC _BorUnwind_RegisterIPLookup(
	_Unwind_IPLookupFn 	LookupFn,
	_Unwind_codePtr 	Start,
	_Unwind_codePtr 	End,
	void *				Context,
	_Unwind_codePtr		GOT)
{
	LookupList *	NewEntry;

#if	DEBUG_CODE_ON
	{
		static	initialized = 0;
		if	(!initialized)
		{
			DEBUG_INIT();
			initialized = 1;
		}
	}
#endif

//DumpStatistics(Context);

	NewEntry = (LookupList *) malloc(sizeof(*NewEntry));
	if	(!NewEntry)
		return 0;

	NewEntry->Func = LookupFn;
	NewEntry->Start = Start;
	NewEntry->End = End;
	NewEntry->GOT = GOT;
	NewEntry->Context = Context;

	if	(LookList)
	{
		NewEntry->Next = LookList;
		NewEntry->Prev = LookList->Prev;
		LookList = NewEntry;
		LookList->Prev->Next = NewEntry;
		LookList->Next->Prev = NewEntry;
	}
	else
	{
		LookList = NewEntry;
		NewEntry->Next = NewEntry;
		NewEntry->Prev = NewEntry;
	}
	return 1;
}

UNWIND_PUBLIC void UNWIND_CC _BorUnwind_UnregisterIPLookup(_Unwind_codePtr Start)
{
	LookupList *	Temp;

	assert(LookList);
	Temp = LookList;
	do
	{
		if	(Temp->Start == Start)
		{
			if	(LookList == Temp)
			{
				LookList = Temp->Next;
				if	(LookList == Temp)
					LookList = NULL;
			}
			Temp->Next->Prev = Temp->Prev;
			Temp->Prev->Next = Temp->Next;
			free(Temp);
			break;
		}
		Temp = Temp->Next;
	}	while	(Temp != LookList);
}

#ifdef	__IA32__
#define MAX_EXTENSION     512
typedef unsigned long	ULONG;
typedef unsigned char	UBYTE;
#define SIZE_FP_REGISTERS      80
typedef struct FP_SAVE
{
    ULONG   Control;
    ULONG   Status;
    ULONG   Tag;
    ULONG   ErrorOff;
    ULONG   ErrorSel;
    ULONG   DataOff;
    ULONG   DataSel;
    UBYTE   Registers[SIZE_FP_REGISTERS];
    ULONG   Cr0NpxState;
}	FP_SAVE;

/*
	Don't rearrange _Unwind_Context arbitrarily!  There's ASM code here which
	manipulates this directly.
*/
struct	_Unwind_Context
{
    ULONG   Eip;
    ULONG   Ebp;
    ULONG   Esp;
    ULONG   Ebx;
    ULONG   Dr0;
    ULONG   Dr1;
    ULONG   Dr2;
    ULONG   Dr3;
    ULONG   Dr6;
    ULONG   Dr7;
    FP_SAVE FloatSave;
    ULONG   SegGs;
    ULONG   SegFs;
    ULONG   SegEs;
    ULONG   SegDs;
    ULONG   Edi;
    ULONG   Esi;
    ULONG   Edx;
    ULONG   Ecx;
    ULONG   Eax;
    ULONG   SegCs;
    ULONG   EFlags;
    ULONG   SegSs;
    UBYTE   ExtendedRegisters[MAX_EXTENSION];
};

static int	CheckAlternatives(struct _Unwind_Context *CPUContext, int DestroySignatures)
{
	_Unwind_codePtr *	StackPtr;

	/*
		We couldn't find the given address in the PC map.  So we try looking
		up the stack for the magic markers that indicate we made a foreign
		call.  If we find one, we ditch all the stack up to that point, reset
		the EIP in the context, and restart the search from there.  If we
		don't find one, we bail out completely.
	*/
	DEBUG_PRINT("unwind.search.image", ("   Checking alternatives\n"));
	for	(StackPtr = (_Unwind_codePtr *)CPUContext->Esp;
		 *StackPtr != UFI_TOPOFSTACK;
		 StackPtr++)
	{
		if	(*StackPtr == UFI_STACK_MARKER)
		{
			unsigned			ParamBytes;
			_Unwind_codePtr		RetAddr;
			ForeignCallInfo *	Info;

			if	(DestroySignatures)
			{
				DEBUG_PRINT("unwind.search.image", ("   Found marker, destroying signature\n"));
				*StackPtr = *StackPtr + 1;
			}

			Info = (ForeignCallInfo *)StackPtr;

			ParamBytes = Info->ParamBytes & ~0x80000000;
			DEBUG_PRINT("unwind.search.image", ("   Found marker, Info->ParamBytes = %08x\n", Info->ParamBytes));
			/*
				If the parameter count is not DWORD aligned, there are two possibilities.
				Either the stack has been bashed (we're hosed), or the signature is not really
				one of ours.  In either case, we really don't want to use the signature as
				the basis for a conservative unwind, so we bail out.
			*/
			// Better be dword aligned!
			if	(Info->ParamBytes & 3)
				return 0;
//			assert(!(Info->ParamBytes & 3));
			RetAddr = *(_Unwind_codePtr *)(((char *)Info) - ParamBytes - 4);
			CPUContext->Eip = RetAddr;
			CPUContext->Ebp = Info->SavedEBP;
			if	(Info->ParamBytes & 0x80000000)
				CPUContext->Esp = ((ULONG)Info);
			else
				CPUContext->Esp = ((ULONG)Info) - ParamBytes;
			DEBUG_PRINT("unwind.search.image", ("   Found marker, RA(0x%08x) EBP(0x%08x) ESP(0x%08x)\n",
				CPUContext->Eip, CPUContext->Ebp, CPUContext->Esp));
			return 1;
		}
	}
	DEBUG_PRINT("unwind.search.image", ("   No markers found\n"));
	return 0;
}

static __personality_routine	FindPersonality(
	struct _Unwind_Context *	CPUContext,
	_Unwind_codePtr 			IP,
	void **						pImplementationData,
	int							DestroySignatures)
{
	LookupList *			Temp;
	__personality_routine	Personality;

	DEBUG_PRINT("unwind.search.image", ("   Looking for %x\n", IP));
restart:
	for	(;;)
	{
		if	(LookList)
		{
			Temp = LookList;
			do
			{
				DEBUG_PRINT("unwind.search.verbose", ("   Looking for %x in (%x, %x)\n",
													IP, Temp->Start, Temp->End));
				if	(IP >= Temp->Start && IP < Temp->End)
				{
					for	(;;)
					{
						Personality = (Temp->Func)(IP, Temp->Context, pImplementationData);
						if	(Personality)
						{
							LookList = Temp;
							DEBUG_PRINT("unwind.search.verbose", ("   Found personality %p\n", Personality));
							return Personality;
						}
						if	(!CPUContext || !CheckAlternatives(CPUContext, DestroySignatures))
							return NULL;
						IP = CPUContext->Eip;
						goto restart;
					}
				}
				Temp = Temp->Next;
			}	while	(Temp != LookList);
		}
		if	(!CPUContext || !CheckAlternatives(CPUContext, DestroySignatures))
			break;
		IP = CPUContext->Eip;
	}
	return NULL;
}

UNWIND_PUBLIC int _BorUnwind_AddressIsInPCMap(_Unwind_codePtr Addr)
{
	if	(FindPersonality(NULL, Addr, NULL, 0))
	{
		DEBUG_PRINT("unwind.control.phases", ("Address %p is in PC map\n", Addr));
		return 1;
	}
	DEBUG_PRINT("unwind.control.phases", ("Address %p is not in PC map\n", Addr));
	return 0;
}

UNWIND_PUBLIC _Unwind_codePtr UNWIND_CC _BorUnwind_FindGOT(_Unwind_codePtr IP)
{
	LookupList *			Temp;

	if	(LookList)
	{
		Temp = LookList;
		do
		{
			if	(IP >= Temp->Start && IP < Temp->End)
			{
				DEBUG_PRINT("unwind.control.phases", ("GOT for EIP %p is %p\n", IP, Temp->GOT));
				return Temp->GOT;
			}
			Temp = Temp->Next;
		}	while	(Temp != LookList);
	}
	return 0;
}

#ifdef	__WIN32__

static __declspec(naked)	__cdecl void GetContext(struct _Unwind_Context *Context)
{
	asm
	{
		push	eax
		push	ecx
		mov		eax, dword ptr [esp + 12]	// Get Context
		mov		ecx, dword ptr [ebp + 4]	// EIP in foreign RTL
		mov		dword ptr [eax], ecx		// save EIP in context
		mov		ecx, [ebp]					// EBP in foreign RTL
		mov		dword ptr [eax + 4], ecx	// save EBP in context
		mov		ecx, ebp					// ESP of parent, at the frame
		add		ecx, 4 + 4					// account for EBP, and RA to foreign RTL
		mov		dword ptr [eax + 8], ecx	// save calced ESP in context
		pop		ecx
		pop		eax
		ret
	}
#pragma message ("finish out the full context")
}
#else
static void GetContext(struct _Unwind_Context *Context)
{
	asm("movl (%ebp), %eax");			// get parent EBP
	asm("movl 4(%eax), %edx");			// RA from parent to foreign RTL
	asm("movl 8(%ebp), %ecx");			// ECX gets ptr to context
	asm("movl %edx, (%ecx)");			// save user EIP
	asm("movl %eax, %edx");				// Calculate ESP in foreign RTL
	asm("add $8, %edx");
	asm("movl (%eax), %eax");			// get foreign EBP
	asm("movl %eax, 4(%ecx)");			// save foreign EBP in context
	asm("movl %edx, 8(%ecx)");			// save foreign ESP in context
#pragma message ("finish out the full context for completeness")
}
#endif
#endif

UNWIND_PUBLIC _Unwind_RegVal UNWIND_CC _BorUnwind_GetIP(const struct _Unwind_Context *Context)
{
	return Context->Eip;
}

UNWIND_PUBLIC void UNWIND_CC _BorUnwind_SetIP(struct _Unwind_Context *Context, _Unwind_RegVal NewValue)
{
	Context->Eip = NewValue;
}

UNWIND_PUBLIC _Unwind_RegVal UNWIND_CC _BorUnwind_GetGR(const struct _Unwind_Context *Context, int Index)
{
	switch	(Index)
	{
	case	_UW_REG_EBP:
		return Context->Ebp;
		break;
	case	_UW_REG_ESP:
		return Context->Esp;
		break;
	case	_UW_REG_EBX:
		return Context->Ebx;
		break;
	default:
		assert(!"Not implemented");
	}
	return 0;
}

UNWIND_PUBLIC void UNWIND_CC _BorUnwind_SetGR(
	struct _Unwind_Context *	Context,
	int 						Index,
	_Unwind_RegVal 				NewValue)
{
	switch	(Index)
	{
	case	_UW_REG_EBP:
		Context->Ebp = NewValue;
		break;
	case	_UW_REG_ESP:
		Context->Esp = NewValue;
		break;
	case	_UW_REG_EBX:
		Context->Ebx = NewValue;
		break;
	default:
		assert(!"Not implemented");
	}
}

static	void	InstallAndResume(
	_Unwind_Context *	Context,			// ebp + 8
	_Unwind_Exception *exception_object,	//ebp + 12
	_Unwind_RegVal CatchIP,
	__personality_routine	Personality)
{
#pragma message ("Use the exception_object correctly in the next release")
//	if	(exception_object)
//	{
//	exception_object->private_1 = CatchIP;
//	exception_object->private_2 = (_Unwind_Word)Personality;
//	}
#ifdef	__IA32__
	{
		unsigned int	TargEIP;
		unsigned int	TargEBP;
		unsigned int	TargESP;
		unsigned int	TargEBX;
		TargEIP = Context->Eip;
		TargEBP = Context->Ebp;
		TargESP = Context->Esp;
		TargEBX = Context->Ebx;
#ifdef	__WIN32__
		asm
		{
			mov		eax, exception_object
			mov		edx, TargEIP
			mov		ecx, TargESP
			mov		ebx, TargEBX
			mov		ebp, TargEBP
			mov		esp, ecx
			jmp		edx
		}
#else
		asm("movl 8(%ebp), %ecx");		// ecx -> context
		asm("movl 12(%ebp), %eax");		// eax -> exception_object
		asm("movl (%ecx), %edx");		// edx -> target EIP
		asm("movl 12(%ecx), %ebx");		// ebx -> target EBX
		asm("movl 4(%ecx), %ebp");		// restore ebp
		asm("movl 8(%ecx), %esp");		// restore esp
		asm("jmp *%edx");
#endif
	}
#endif
}

UNWIND_PUBLIC void UNWIND_CC _BorUnwind_Resume(_Unwind_Exception *exception_object)
{
	assert(!"Not implemented");
}

#ifdef	__linux__
static _Unwind_codePtr GetFSChain(void)
{
	asm("xor %eax, %eax");
	asm("mov %fs:(%eax), %eax");
}

/*
	Wine will install an exception change threaded through FS:[0], a'la Win32.
	If there is such a chain, we need to respect it when deciding whether there is
	a better (closer) Delphi exception handler in place to deal with exceptions.
	ClosestDelphiHandler does this comparison.
*/
UNWIND_PUBLIC _Unwind_codePtr UNWIND_CC _BorUnwind_ClosestDelphiHandler(
	struct sigcontext *SContext)
{
	__personality_routine	Personality;
	void *					PersonalityData;
	_Unwind_Context			Context;
	_Unwind_RegVal			CatchIP;
	_Unwind_Reason_Code		Res;

	DEBUG_PRINT("unwind.control.phases", ("Looking for closest handler\n"));
	
	Context.Ebp = SContext->ebp;
	Context.Esp = SContext->esp;
	Context.Eip = SContext->eip;
	Context.Eax = SContext->eax;
	Context.Ebx = SContext->ebx;
	Context.Ecx = SContext->ecx;
	Context.Edx = SContext->edx;
	Context.Edi = SContext->edi;
	Context.Esi = SContext->esi;

	for	(;;)
	{
		CatchIP = _BorUnwind_GetIP(&Context);
		Personality = FindPersonality(&Context, CatchIP, &PersonalityData, 0);
		if	(!Personality)
		{
			DEBUG_PRINT("unwind.control.phases", (" Search closest handler error - no personality\n"));
			return 0;
		}
		Res = (Personality)(_UNWIND_VERSION,
						    _UA_SEARCH_PHASE, 
						    0,
						    &Context,
						    PersonalityData);
		if	(Res == _URC_HANDLER_FOUND)
			break;
		if	(Res != _URC_CONTINUE_UNWIND)
		{
			DEBUG_PRINT("unwind.control.phases", (" Search closest handler error - strange personality return\n"));
			return 0;
		}
	}
	return (Context.Esp < GetFSChain()) ? 1 : 0;
}
#endif

UNWIND_PUBLIC _Unwind_Reason_Code UNWIND_CC _BorUnwind_RaiseException(
	_Unwind_Exception *exception_object)
{
	__personality_routine	Personality;
	void *					PersonalityData;
	_Unwind_Context			Context;
	_Unwind_Context			SaveContext;
	_Unwind_RegVal			CatchIP;
	_Unwind_Reason_Code		Res;

	GetContext(&Context);
	SaveContext = Context;

	DEBUG_PRINT("unwind.control.phases", ("RaiseException\n"));
	DEBUG_PRINT("unwind.control.phases", ("Search Phase:\n"));

	for	(;;)
	{
		CatchIP = _BorUnwind_GetIP(&Context);
		Personality = FindPersonality(&Context, CatchIP, &PersonalityData, 0);
		if	(!Personality)
		{
			DEBUG_PRINT("unwind.control.phases", (" Search fatal phase 1 error - no personality\n"));
			return _URC_FATAL_PHASE1_ERROR;
		}
		Res = (Personality)(_UNWIND_VERSION,
						    _UA_SEARCH_PHASE, 
						    exception_object,
						    &Context,
						    PersonalityData);
		if	(Res == _URC_HANDLER_FOUND)
			break;
		if	(Res != _URC_CONTINUE_UNWIND)
		{
			DEBUG_PRINT("unwind.control.phases", (" Search fatal phase 1 error - strange personality return\n"));
			return _URC_FATAL_PHASE1_ERROR;
		}
	}

	/*
		Found a handler.  Restart, telling frames to clean themselves up.
	*/
	
	DEBUG_PRINT("unwind.control.phases", ("Cleanup Phase:\n"));
	while	(_BorUnwind_GetIP(&SaveContext) != CatchIP)
	{
		Personality = FindPersonality(&SaveContext, _BorUnwind_GetIP(&SaveContext), &PersonalityData, 1);
		if	(!Personality)
		{
			return _URC_FATAL_PHASE2_ERROR;
		}
		Res = (Personality)(_UNWIND_VERSION,
						    _UA_CLEANUP_PHASE,
						    exception_object,
						    &SaveContext,
						    PersonalityData);
		if	(Res == _URC_INSTALL_CONTEXT)
		{
			/*
				Hoo boy.  The phase 2 handler wants to be running without a net.
				OK.  Install the context, after saving off some data that we'll
				need on a resume.
			*/
			InstallAndResume(&SaveContext, exception_object, CatchIP, Personality);
			// Never returns
			assert(!"Never should return");
#pragma message("How is Resume supposed to work now?")
		}
		if	(Res != _URC_CONTINUE_UNWIND)
			return _URC_FATAL_PHASE2_ERROR;
	}
	/*
		We should now be at the frame that wanted the exception.
	*/
	Personality = FindPersonality(&SaveContext, _BorUnwind_GetIP(&SaveContext), &PersonalityData, 1);
	if	(!Personality)
		return _URC_FATAL_PHASE2_ERROR;
	Res = (Personality)(_UNWIND_VERSION,
					    _UA_CLEANUP_PHASE | _UA_HANDLER_FRAME,
					    exception_object,
					    &SaveContext,
					    PersonalityData);
	if	(Res != _URC_INSTALL_CONTEXT)
		return _URC_FATAL_PHASE2_ERROR;

	/*
		We've come to the point where the personality routine for the catch
		frame has run its cleanup phase.  We require it to install a context,
		and continue execution at this point.  This is currently not resumeable.
	*/
#pragma message ("Should this be resumeable?")
	InstallAndResume(&SaveContext, exception_object, 0, NULL);
	assert(!"Never should get here");
	return _URC_NO_REASON;
}

