/* **************************************************************************
 *                                                                          *
 * unwind.h --- Exception stack unwind routines as specified by the C++ ABI *
 *   for IA-64:Exception Handling specification                             *
 *                                                                          *
 * Copyright (c) 2001 Borland Software Corporation.                         *
 *                                                                          *
 * This file may be distributed and/or modified under the terms of the GNU  *
 * Lesser General Public License (LGPL) version 2.1 as published by the     *
 * Free Software Foundation and appearing at                                *
 * http://www.borland.com/kylix/lgpl.html.                                  *
 *                                                                          *
 * Licensees holding a valid Borland No-Nonsense License for this Software  *
 * may  use this file in accordance with such license, which appears in the *
 * file license.txt that came with this Software.                           *
 *                                                                          *
 * *************************************************************************/

#ifndef	UNWIND_H
#define	UNWIND_H

/****************************************************************************

	This implementation is based of of the specs given at the URL below.
	
	http://reality.sgi.com/dehnert_engr/cxx/abi-eh.html
	
	There is a PDF file of the same thing.  Look here:
	http://reality.sgi.com/dehnert_engr/cxx/cxx-summary.html

 ****************************************************************************/

#ifdef	__cplusplus
extern "C" {
#endif

/****************************************************************************
 Pick one:
 ****************************************************************************/
	#define	__IA32__
//	#define	__IA64__

/****************************************************************************
 Delphi specific:
 ****************************************************************************/
	#define	UNWIND_DELPHI_SUPPORT

/****************************************************************************
 Processor specific definitions
 ****************************************************************************/
#ifdef	__IA32__
typedef	unsigned int	_Unwind_codePtr;
typedef	unsigned int	_Unwind_Word;

// Register value
// what about fp regs?  Do 'em in parts?
typedef unsigned int	_Unwind_RegVal;

#ifdef	__WIN32__
#define	UNWIND_CC		__cdecl
#define UNWIND_PUBLIC	__declspec(dllexport)
#else
#define	UNWIND_CC
#define UNWIND_PUBLIC
#endif
#endif

#ifdef	__IA64__
typedef	uint64			_Unwind_codePtr;
typedef	uint64			_Unwind_Word;

// Register value
typedef uint64			_Unwind_RegVal;
#endif


/****************************************************************************
 Version of this interface, as passed to personality routines
 ****************************************************************************/
#define	_UNWIND_VERSION	1

/****************************************************************************
 Return codes from personality routines
 ****************************************************************************/
typedef	enum
{
	_URC_NO_REASON = 0,
	_URC_FOREIGN_EXCEPTION_CAUGHT = 1,
	_URC_FATAL_PHASE2_ERROR = 2,
	_URC_FATAL_PHASE1_ERROR = 3,
	_URC_NORMAL_STOP = 4,
	_URC_END_OF_STACK = 5,
	_URC_HANDLER_FOUND = 6,
	_URC_INSTALL_CONTEXT = 7,
	_URC_CONTINUE_UNWIND = 8
}	_Unwind_Reason_Code;

/****************************************************************************
 Modes for personality routines
 ****************************************************************************/
#ifdef __cplusplus
typedef int	_Unwind_Action;
const _Unwind_Action _UA_SEARCH_PHASE = 1;
const _Unwind_Action _UA_CLEANUP_PHASE = 2;
const _Unwind_Action _UA_HANDLER_FRAME = 4;
const _Unwind_Action _UA_FORCE_UNWIND = 8;
#else
typedef enum
{
	_UA_SEARCH_PHASE = 1,
	_UA_CLEANUP_PHASE = 2,
	_UA_HANDLER_FRAME = 4,
	_UA_FORCE_UNWIND = 8
}	_Unwind_Action;
#endif

/****************************************************************************
 Opaque processor context structure.  Access through context APIs.
 ****************************************************************************/
typedef struct	_Unwind_Context	_Unwind_Context;

typedef struct _Unwind_Exception _Unwind_Exception;
    
/****************************************************************************
 Function to be used to cleanup an arbitrary exception object, if you
 need to destroy the object that has been thrown.
 ****************************************************************************/
typedef void (*_Unwind_Exception_Cleanup_Fn)(_Unwind_Reason_Code reason, _Unwind_Exception *Exc);

/****************************************************************************
 Exception object - this is what you throw
 ****************************************************************************/
struct	_Unwind_Exception
{
	_Unwind_Word					exception_class;
	_Unwind_Exception_Cleanup_Fn	exception_cleanup;
	_Unwind_Word					private_1;
	_Unwind_Word					private_2;
};

/****************************************************************************
 Personality routine - called to unwind frames at exception time, and to
 handle language implementation specifics for tearing down frames, like
 object destruction.
 ****************************************************************************/
typedef	_Unwind_Reason_Code UNWIND_CC (*__personality_routine)(
	int version,
	_Unwind_Action actions,
//	_Unwind_Word exceptionClass,
	_Unwind_Exception *exceptionObject,
	_Unwind_Context *context,
	void *ImplementationData);

/****************************************************************************
 PC Map lookup routines.  Not currently in the spec.
 ****************************************************************************/
typedef __personality_routine UNWIND_CC (*_Unwind_IPLookupFn)(_Unwind_codePtr, void *Context, void **pImplementationData);

UNWIND_PUBLIC int UNWIND_CC _BorUnwind_RegisterIPLookup(
	_Unwind_IPLookupFn 	LookupFn,
	_Unwind_codePtr 	Start,
	_Unwind_codePtr 	End,
	void *				Context,
	_Unwind_codePtr 	GOT);

/*
	Registers a lookup function for finding a personality routine and implementation data
	to be handed to the personality routine for unwinding support, given an IP value.  This
	is so that the language implementors can have control over the format of their
	PC maps.  Returns 0 on failure, non-zero on success.
*/
UNWIND_PUBLIC void UNWIND_CC _BorUnwind_UnregisterIPLookup(_Unwind_codePtr Start);

/****************************************************************************
 Exception raising.  This is how you raise an exception from a given language.
 ****************************************************************************/
UNWIND_PUBLIC _Unwind_Reason_Code UNWIND_CC _BorUnwind_RaiseException(
	_Unwind_Exception *	exception_object);

/****************************************************************************
 Context manipulation APIs
 ****************************************************************************/
#ifdef	__IA32__
/*
	__IA32__ specific register constants to pass to Get/SetGR:
*/
#define	_UW_REG_EBP	0
#define	_UW_REG_ESP	1
#define	_UW_REG_EBX	2
#endif

UNWIND_PUBLIC _Unwind_RegVal UNWIND_CC _BorUnwind_GetGR(
	const struct _Unwind_Context *	Context,
	int 							Index);

UNWIND_PUBLIC void UNWIND_CC _BorUnwind_SetGR(
	struct _Unwind_Context *	Context,
	int							Index,
	_Unwind_RegVal				NewValue);

// Don't know about these.  I think maybe they should be folded back into Get/SetGR
// and add platform specific consts.
UNWIND_PUBLIC _Unwind_RegVal UNWIND_CC _BorUnwind_GetIP(
	const struct _Unwind_Context *	Context);

UNWIND_PUBLIC void UNWIND_CC _BorUnwind_SetIP(
	struct _Unwind_Context *	Context,
	_Unwind_RegVal 				NewValue);

UNWIND_PUBLIC void UNWIND_CC _BorUnwind_Resume(_Unwind_Exception *exception_object);

UNWIND_PUBLIC int _BorUnwind_AddressIsInPCMap(_Unwind_codePtr Addr);

UNWIND_PUBLIC _Unwind_codePtr UNWIND_CC _BorUnwind_FindGOT(_Unwind_codePtr IP);
UNWIND_PUBLIC void	UNWIND_CC _BorUnwind_CrawlStack(void);

#define	UFI_TOPOFSTACK		0xBE00EF00
#define	UFI_STACK_MARKER	0xBEEFFEED  // emitted by the Delphi compiler in generated code

typedef	struct	ForeignCallInfo
{
	unsigned int	Signature;
	unsigned int	SavedEBP;
	unsigned int	ParamBytes;
}	ForeignCallInfo;

#ifdef	__cplusplus
}
#endif

#endif

/*
Notes:

__personality_routine changed to take implementation data, and not an exception class.

_URC_CONTINUE_UNWIND added to enum.  Wasn't in the doc.

Should an exception be resumeable after a catch frame catches it?
*/

