/* ************************************************************************* *
 *                                                                           *
 * unwindfi.h --- Unwind Frame Info.                                         *
 *                                                                           *
 * Copyright (c) 2000,2001 Borland Software Corporation                      *
 *                                                                           *
 * This file may be distributed under the terms of the GNU General Public    *
 * License (GPL) version 2 as published by the Free Software Foundation and  *
 * appearing at http://www.borland.com/kylix/gpl.html.                       *
 *                                                                           *
 * Licensees holding a valid Borland No-Nonsense License for this Software   *
 * may use this file in accordance with such license, which appears in the   *
 * file license.txt that came with this software.                            *
 *                                                                           *
 * This file includes the descriptors for the Delphi/Kylix PC map. plus the  *
 * frame unwind info and local PC map. Much of the data structures and       *
 * constants here are used by the compiler/linker when emitting PC map and   *
 * unwind information. Changes here require a rebuild of the compiler and    *
 * all the dependant files.                                                  *
 *                                                                           *
 * ************************************************************************ */

#ifndef	UNWINDFI_H
#define	UNWINDFI_H

#ifdef __cplusplus
extern "C" {
#endif

#define	UFI_F_REG_EBX		0x01	/* EBX is saved */
#define	UFI_F_REG_ESI		0x02	/* ESI is saved */
#define	UFI_F_REG_EDI		0x04	/* EDI is saved */
#define	UFI_F_REG_EBP		0x08	/* EBP is saved */
#define UFI_F_HASFRAME		0x10	/* proc has an EBP frame */
#define UFI_F_HASRANGES		0x20	/* proc has any protected ranges at all */
#define UFI_F_SROFFS		0x40	/* saved regs offset was stored */
#define UFI_F_CALLEEPOP		0x80	/* proc has callee pop amount */
#define UFI_F_NESTEDHANDLER 0x08	/* overloading REG_EBP - means that unwind info is present */

/*
	If you change this structure, you must change ReverseFrameInfo in objout.c!
*/
typedef	struct ExceptionFrameInfo
{
	unsigned char	Flags;				/* UFI_F_* above */
	unsigned int	SavedRegsEBPOffs;
	unsigned int	OffsetToPush;
	unsigned int	RangeCount;
	unsigned int	UnwindInfoSize;
	unsigned int	ProcLen;
	unsigned int	CalleePopAmount;
}	ExceptionFrameInfo;

typedef	struct ExceptionRangeOffsets
{
	int	Start;
	int	End;
	int	Handler;
}	ExceptionRangeOffsets;


typedef	struct	PCMapUnit
{
	long	Start;
	long	End;
	int		ProcCount;
	long	ProcInfoOffs;
}	PCMapUnit;

typedef	struct	PCMapProc
{
	long	Start;
	long	End;
}	PCMapProc;

/*
	Borland Delphi Exceptions version 0x00.0x08
*/
#define	PCMAPHEADER_SIGNATURE	0xFBDE0008

typedef	struct	PCMapHeader
{
	unsigned long	Signature;
	long			CodeSegmentSize;
	int				UnitCount;
}	PCMapHeader;

#ifdef __cplusplus
}
#endif
#endif
